/**
 ** @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: Release 1.1 $date:  June 17, 1999$
 ** (c) 1999 by Ben Yanbin Zhao and the Regents of the University of California
 **
 ** This is the utility support class for the SET server.
 ** An ongoing progress homepage of the project can be found at 
 ** http://www.cs.berkeley.edu/~ravenben/xset
 **/

package xset;

import java.util.*;
import java.io.*;
import java.security.*;
import org.w3c.dom.*;
import com.ibm.xml.parsers.DOMParser;
import org.xml.sax.InputSource;

public class SETutils {

  /************************************
   ** Vector queryMax(Vector, String) ***
   *************************************
   ** Method that returns the desired document or tags that represents
   ** the maximum value in a given tag. Just a wrapper for queryMaxMin
   ** @param xmlVector The vector of XML documents to be compared
   ** @param rangeTag The string name of the tag the max is executed on
   ** @return vector of simple xml documents as strings
   **/
  public static Vector queryMax(Vector xmlVector, String rangeTag) {
      return queryMaxMin(xmlVector, rangeTag, 1);
  }

  /************************************
   ** Vector queryMin(Vector, String) ***
   *************************************
   ** Method that returns the desired document or tags that represents
   ** the minimum value in a given tag. Just a wrapper for queryMaxMin
   ** @param xmlVector The vector of XML documents to be compared
   ** @param rangeTag The string name of the tag the min is executed on
   ** @return vector of simple xml documents as strings
   **/
  public static Vector queryMin(Vector xmlVector, String rangeTag) {
      return queryMaxMin(xmlVector, rangeTag, -1);
  }

  /********************************************
   ** Vector queryMaxMin(Vector, String, int) ***  
   *********************************************
   ** Method that performs selection on max or min criteria on a tag
   ** Internal method that is wrapped as queryMax and queryMin
   ** @param xmlVector The vector of XML documents to be compared
   ** @param rangeTag The string name of the tag the min is executed on
   ** @param maxmin The int that tells the method whether to perform max or min
   ** @return vector of simple xml documents as strings
   **/
  private static Vector queryMaxMin(Vector xmlVector, String rangeTag,
				    int maxmin) {
      if (xmlVector == null || rangeTag == null || maxmin == 0)
	  return new Vector();
      DOMParser parser = new DOMParser();
      Enumeration docs = xmlVector.elements();
      String currXML, currMinMax=null, tmp;
      Document d;
      ByteArrayInputStream bais;
      NodeList nl;
      Vector tempV, results = new Vector(), v=new Vector(); 
      Element elem;
      while (docs.hasMoreElements()) {
	  currXML = (String) docs.nextElement();
	  bais = new ByteArrayInputStream(currXML.getBytes());
	  try {
	      parser.parse(new InputSource(bais));
	  } catch (Exception e) {
	      e.printStackTrace(System.err);
	  }
	  d = parser.getDocument();
	  nl = d.getElementsByTagName(rangeTag);
	  parser.reset();
	  v.removeAllElements();
	  for(int i=0, j=nl.getLength(); i < j; i++) {
	      // FOR each tag found of this tagname, DO:
	      v.addElement(getText(nl.item(i)));
	  } //for
	  tmp = stringMaxMin(v, maxmin);
	  if (currMinMax == null) {
	      currMinMax = tmp;
	      continue;
	  }
	  int i = currMinMax.compareTo(tmp) * maxmin;
	  //take the product of the 2 min max values
	  if (i < 0) { 
	      //negative means tmp better than optimal
	      //we reset currMinMax, clear the results, and add the curr XML
	      currMinMax = tmp;
	      results.removeAllElements();
	      results.addElement(currXML);
	  } else if (i == 0) {
	      //results are same as optimal, we add currXML to results
	      results.addElement(currXML);
	  }
      }
      return results;
  }
    private static String getText(Node n) {
	Node child;
	if (n.getNodeType() == Node.ELEMENT_NODE) {
	    
	    if ((child = n.getFirstChild()) != null) {
		return child.getNodeValue();
	    }
	} else if (n.getNodeType() == Node.TEXT_NODE)
	    return n.getNodeValue();
	return null;
    }
  /*************************************
   ** String stringMaxMin(Vector, int) ***
   **************************************
   ** Method that rerturns the min or max string from a vector of strings
   ** @param stringV the vector of strings that we take the max or min from
   ** @param maxmin, indicator of max or min, if <0 min, if >0 max, if 0 null
   ** @return string that is the end result value as the max or min
   **/
  public static String stringMaxMin(Vector stringV, int maxmin) {
      if (stringV == null || stringV.size() == 0) return null;
      //optimize for the common case: most calls from queryMaxMin w/ 1 string
      if (stringV.size() == 1) return (String) stringV.firstElement();
      
      String optimal=null, curr;
      Enumeration e = stringV.elements();
      optimal = (String) e.nextElement();
      if (maxmin < 0) {  //trying to find min
	  while(e.hasMoreElements()) {
	      curr = (String) e.nextElement();
	      if (optimal.compareTo(curr) > 0) optimal = curr;
	  }
      } else if (maxmin > 0) { // trying to find max
	  while(e.hasMoreElements()) {
	      curr = (String) e.nextElement();
	      if (optimal.compareTo(curr) < 0) optimal = curr;
	  }
      } else return null;
      return optimal;
  }

  /**
   ** Method to take an XML4J DOM element and count its number of children
   ** Whitespaces here are not ignored. For that value, do getChildren().size()
   ** @param e The DOM element in question
   ** @return integer that represents the number of children it has
   **/
  public static int numChildren(org.w3c.dom.Element e) {
      org.w3c.dom.Node ptr = null;
      int i;
      if ((ptr = e.getFirstChild()) == null) return 0;
      else i=1;
      while (true) {
	  if ((ptr = ptr.getNextSibling()) != null) i++;
	  else return i;
      }
  }

  /**
   ** Methods that takes an XML4J/DOM element and returns its children,
   ** ignoring whitespaces between tags
   ** @param e The DOM element that we find nonwhitespace children for
   ** @return Vector containing the child DOM elements
   **/
  public static Vector getChildren(org.w3c.dom.Element e) {
      org.w3c.dom.Node a = null, b = null;
      if ((a = e.getFirstChild()) == null) {
	  return null;
      } else {
	  Vector children = new Vector();
	  //Standard parser eliminates comments in parsing, so we don't
	  //worry about it here
	  if (a.getNodeType() != org.w3c.dom.Node.TEXT_NODE) {
	      children.addElement(a);
	  } else if (!ignorable(a)) children.addElement(a);
	  while ((a = a.getNextSibling()) != null) {
	      if (a.getNodeType() != org.w3c.dom.Node.TEXT_NODE) {
		  children.addElement(a);
	      } else {
		  if (!ignorable(a)) children.addElement(a);
	      }
	  }
	  if (children.isEmpty()) return null;
	  return children;
      }
  }
  private static boolean ignorable(org.w3c.dom.Node e) {
      String temp = e.getNodeValue();
      return (temp.trim().length() == 0);
  }

  /**
   ** Simplistic method to print out an element with its immediate children
   **/
  public static void printElement(org.w3c.dom.Element n) { 
      System.out.println("Begin tag: "+n.getNodeName()+"\n");
      System.out.println("Value: "+n.getNodeValue()+"\n");
      Vector children = getChildren(n);
      if (children != null) {
	  Enumeration e = children.elements();
	  int i = 0;
	  org.w3c.dom.Element ec = null;
	  while (e.hasMoreElements()) {
	      System.out.println("child #"+(++i));
	      ec = (org.w3c.dom.Element) e.nextElement();
	      System.out.println("  has children?"+ec.hasChildNodes());
	      System.out.println("  Child Tag: "+ec.getNodeName());
	      System.out.println("  Value: >>"+ec.getNodeValue()+"<<");
	  }
      }
  }
    
  /**
   ** Method that packages up a file into an inputstream
   **/
  public static InputStream parseFile(String f) throws FileNotFoundException {
    InputStream is = null;
    is = new FileInputStream(f);
    return is;
  }
  /**
   ** Method that implements a hashing algorithm (MD5 or SHA-1).  Takes in
   ** a string and returns a string that represented the hashed value.
   ** 
   ** @param inString The input that is to be hashed
   ** @param alg The algorithm to be used, either "MD5" or "SHA-1"
   ** @exception NoSuchAlgorithmException thrown when algorithm is unknown
   ** @return the hashed value of inString
   **/
  static public String hasher(String inString, String alg) 
       throws NoSuchAlgorithmException {
	 if ((!alg.equals("MD5")) && !alg.equals("SHA-1")) {
	   throw new NoSuchAlgorithmException();
	 }
	 MessageDigest md = MessageDigest.getInstance(alg);
	 byte [] inarray = null, outarray = null;
	 String output = null;
	 inarray = inString.getBytes();
	 md.update(inarray);
	 outarray = md.digest();
	 output = new String(outarray);
	 return output;
  }

  static public String hasher(org.w3c.dom.Document d, String alg) 
       throws NoSuchAlgorithmException {
	 if ((!alg.equals("MD5")) && !alg.equals("SHA-1")) {
	   throw new NoSuchAlgorithmException();
	 }
	 MessageDigest md = MessageDigest.getInstance(alg);
	 byte [] inarray = null, outarray = null;
	 String output = null;
	 inarray = XMLHelper.DocumentToString(d).getBytes();
	 md.update(inarray);
	 outarray = md.digest();
	 output = new String(outarray);
	 return output;
  }
  
}
